/*
 * Copyright (c) 2005
 *    Christian Biere <christianbiere@gmx.de> All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the authors nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include "filter.h"
#include "nettools.h"

static int
cmp_addr_filter_entry(const void *p, const void *q)
{
  uint32_t a = *(const uint32_t *) p, b = *(const uint32_t *) q;
  return a == b ? 0 : (a > b ? 1 : -1);
}

addr_filter_t *
addr_filter_load(const char *filename)
{
  FILE *f;
  unsigned int line = 0;
  addr_filter_t *af;
  size_t i;
  char buf[4096];
 
  af = calloc(1, sizeof *af);
  if (!af) {
    CRIT("Out of memory");
    return NULL;
  }
  
  for (i = 0; i < ARRAY_LEN(af->nets); ++i) {
    af->nets[i].addrs = NULL;
    af->nets[i].n = 0;
  }
  
  f = safer_fopen(filename, SAFER_FOPEN_RD);
  if (!f) {
    WARN("could not open \"%s\": %s", filename, compat_strerror(errno));
    return NULL;
  }
  
  for (;;) {
    uint32_t addr, mask;
    uint64_t v;
    int c, ret, tokens;
    char *ep, *p;
    void *q;

    p = buf;
    tokens = tokenline(f, buf, sizeof buf);
    if (tokens < 0) {
      break;
    }
    line++;

    if (tokens < 1) {
      /* empty (line ignored) */
      continue;
    }
    if (*p == '#') {
      /* comment (line ignored) */
      continue;
    }
    
    ret = parse_ipv4_addr(p, &addr, &ep);
    c = *ep;
    p = ++ep;
    if (!ret || (c != '\0' && c != '/')) {
      WARN("Invalid data in line %u; expected IPv4 address (line ignored)",
          line);
      continue;
    }

    if (c == '\0') {
      mask = ~((uint32_t) 0U);
      v = 32;
    } else {
      uint32_t u;
      int error;

      v = parse_uint64(p, &ep, 10, &error);
      if (v < 1 || v > 32 || *ep != '\0') {
        ret = parse_ipv4_addr(p, &mask, &ep);
        if (!ret || 0 == mask || '\0' != *ep) {
          WARN("Invalid data in line %u; expected IPv4 netmask (line ignored)",
            line);
          continue;
        }
        
        v = 32;
        for (u = ntohl(mask); !(u & 1); u >>= 1) {
          v--;
        }
        u = 0xffffffffU << (32 - v);
        if (mask != htonl(u)) {
          WARN("Strange netmask in line %u; expected IPv4 netmask "
              "(line ignored)", line);
          continue;
        }
      } else {
        u = 0xffffffffU << (32 - v);
        mask = htonl(u);
      }
    }

    assert(mask != 0);
    v = 32 - v;
    assert((int) v >= 0 && v < ARRAY_LEN(af->nets));

    q = realloc(af->nets[v].addrs,
            (af->nets[v].n + 1) * sizeof af->nets[v].addrs[0]);
    if (!q) {
      CRIT("Out of memory");
      break;
    }

    {
      char addr_buf[IPV4_ADDR_BUFLEN];
      char mask_buf[IPV4_ADDR_BUFLEN];
       
      print_ipv4_addr(addr_buf, sizeof addr_buf, addr);
      print_ipv4_addr(mask_buf, sizeof mask_buf, mask);
#if 0
      DBUG("Adding %s/%u (mask=%s) to filter",
          addr_buf, (unsigned) (32 - v), mask_buf);
#endif
    }

    af->nets[v].addrs = q;
    af->nets[v].addrs[af->nets[v].n] = addr & mask;
    af->nets[v].n++;
  }
  fclose(f);

  for (i = 0; i < ARRAY_LEN(af->nets); ++i) {
    if (!af->nets[i].addrs)
      continue;
    
    qsort(af->nets[i].addrs, af->nets[i].n, sizeof af->nets[i].addrs[0],
      cmp_addr_filter_entry);
  }
  
  return af;
}

bool
addr_filter_match(const addr_filter_t *af, net_addr_t addr)
{
  in_addr_t ip;
  size_t i;

  /* XXX: Add IPv6 support */
  if (!net_addr_is_ipv4_mapped(addr))
    return false;
  ip = net_addr_ipv4(addr);
  
  for (i = 0; i < ARRAY_LEN(af->nets); ++i) {
    in_addr_t v;
    void *p;
   
    if (!af->nets[i].addrs)
      continue;
    
    v = ip;
    v &= htonl(((in_addr_t) -1) << i);

    p = bsearch(&v, af->nets[i].addrs, af->nets[i].n,
          sizeof af->nets[i].addrs[0], cmp_addr_filter_entry);
    if (p != NULL)
      return true;
  }

  return false;
}

/* vi: set ai et sts=2 sw=2 cindent: */
