/*
 * Copyright (c) 2006 Christian Biere <christianbiere@gmx.de>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the authors nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include "server.h"

#if defined(USE_POLL)
#undef USE_KQUEUE
#undef USE_EPOLL
#else /* !USE_POLL */
#if defined(HAVE_KQUEUE)
#define USE_KQUEUE
#endif /* HAVE_KQUEUE */
#if defined(HAVE_EPOLL)
#define USE_EPOLL
#endif /* HAVE_EPOLL */
#endif /* USE_POLL */

#ifdef USE_EPOLL
#include <sys/epoll.h>
#endif /* USE_EPOLL*/

#include <poll.h>

/* If still not defined, just define them */
#ifndef POLLRDNORM
#define POLLRDNORM 0
#endif
#ifndef POLLRDBAND
#define POLLRDBAND 0
#endif
#ifndef POLLWRNORM
#define POLLWRNORM 0
#endif
#ifndef POLLWRBAND
#define POLLWRBAND 0
#endif

#include "connection.h"
#include "oop.h"

#define SERVER_CHUNKS 64

/* XXX: "server" is really a poor name for this */

enum server_type {
  SERVER_TYPE_NONE,
  SERVER_TYPE_POLL,
  SERVER_TYPE_KQUEUE,
  SERVER_TYPE_EPOLL
};

typedef struct server_private {
  struct server         vtable;

  uint32_t              magic;
  connection_t          **clients;
  size_t                size_clients;
  size_t                num_clients;
  size_t                num_closed;
  server_periodic_cb_t  periodic_cb;
  unsigned long         timeout;  /* delay for periodic callbacks */
  struct timeval        last_gc;

  enum server_type      server_type;

  void                  *ev_arr; /* array of kevent, pollfd, epoll_event */
  size_t                ev_size; /* size of ev_arr[0] */
  int                   ev_fd;  /* kqueue() or epoll() fd, otherwise -1 */
  
} server_private_t;

typedef union {
  server_t pub;
  server_private_t priv;
} server_union;

static const uint32_t server_magic = MAGIC_CHARS('S', 'E', 'R', 'V');

#define PRIV(x) (((server_union *) (x))->priv)

ATTRIBUTE_SET(server, unsigned long, timeout)

static void
server_set_periodic_cb(server_t *server, server_periodic_cb_t cb)
{
  server_private_t *s = &PRIV(server);

#ifdef USE_KQUEUE
  if (SERVER_TYPE_KQUEUE == s->server_type) {
    static const struct timespec ts;
    struct kevent kev;

    EV_SET(&kev, PTR2UINT(server),
        EVFILT_TIMER, (cb ? EV_ADD : EV_DELETE), 0, s->timeout, 0);
    if (-1 == kevent(s->ev_fd, &kev, 1, 0, 0, &ts)) {
      WARN("kevent() failed: %s", compat_strerror(errno));
    }
  }
#endif /* USE_KQUEUE */
  
  s->periodic_cb = cb;
}

static void
server_destruct(server_t *server)
{
  server_private_t *s = &PRIV(server);

  if (-1 != s->ev_fd) {
    close(s->ev_fd);
    s->ev_fd = -1;
  }
  DO_FREE(s->ev_arr);

  memset(s, 0, sizeof *s);
  DO_FREE(s);
}

static void
server_garbage_collect(server_private_t *s, const struct timeval *now)
{
  bool timeouts;
  size_t i, j;

  assert(s);
  assert(now);
  
  timeouts = 0 != DIFFTIMEVAL(now, &s->last_gc) / 1000000;
  if (!timeouts && 0 == s->num_closed)
    return;

  s->last_gc = *now;

  for (i = 0; i < s->size_clients; i++) {
    connection_t *c;
    unsigned int it;
    int fd;

    assert(i < s->size_clients);
    if (NULL == (c = s->clients[i]))
      continue;

    fd = c->get_fd(c);
    if (timeouts && fd >= 0) {
      it = c->get_maxidle(c);
      if (it) {
        struct timeval tv;
      
        tv = c->get_stamp(c);
        if (DIFFTIMEVAL(now, &tv) / 1000000 >= it) {
          c->event(c, CONN_EV_TIMEOUT, now);
          fd = c->get_fd(c);
          if (-1 == fd)
            s->num_closed++;
        }
      }
    }

    if (fd < 0) {
      assert(fd == -1);
      assert(s->num_closed > 0);
      DESTRUCT(c);
      s->clients[i] = NULL;
      s->num_clients--;
      s->num_closed--;
    }

  }

  j = 0;
  for (i = 0; i < s->size_clients; i++) {
    if (!s->clients[i]) {
      while (j < s->size_clients && !s->clients[j])
        j++;

      if (j >= s->size_clients)
        break;

      s->clients[i] = s->clients[j];
      s->clients[j] = NULL;
    }
  }

}

static void
server_watch_client(server_t *server, connection_t *c, connection_event_t ev)
{
  server_private_t *s = &PRIV(server);
  connection_event_t oev;
  int fd;
  bool r = false, w = false;

  assert(c != NULL);
  assert(s->num_clients > 0);
  fd = c->get_fd(c);
  assert(fd >= 0);

  oev = c->get_eventmask(c);
  if (oev == ev)
    return;
  
  switch (ev) {
  case CONN_EV_READ:
    r = true;
    break;
  case CONN_EV_WRITE:
    w = true;
    break;
  case CONN_EV_NONE:
    break;
  default:
    assert(0);
  }

#ifdef USE_KQUEUE
  if (SERVER_TYPE_KQUEUE == s->server_type) {
    struct kevent kev[2];
    size_t n = 0;

    if (oev != CONN_EV_NONE) {
      assert(oev == CONN_EV_WRITE || oev == CONN_EV_READ);

      EV_SET(&kev[n], fd, (oev == CONN_EV_READ ? EVFILT_READ : EVFILT_WRITE),
          EV_DELETE | EV_DISABLE, 0, 0, PTR_TO_KEVENT_UDATA(c));
      n++;
    }

    if (r || w) {
      EV_SET(&kev[n], fd, (r ? EVFILT_READ : EVFILT_WRITE),
          EV_ADD, 0, 0, PTR_TO_KEVENT_UDATA(c));
      n++;
    }

    if (n != 0) {
      static const struct timespec ts;

      assert(n <= ARRAY_LEN(kev));
      if (-1 == kevent(s->ev_fd, kev, n, NULL, 0, &ts)) {
        WARN("kevent() failed: %s", compat_strerror(errno));
      }
    }
  }
#endif /* USE_KQUEUE */
  
#ifdef USE_EPOLL
  if (SERVER_TYPE_EPOLL == s->server_type) {
    static const struct epoll_event zero_pe;
    struct epoll_event pe;
    int op;

    pe = zero_pe;
    pe.data.ptr = c;
    if (r) {
      pe.events |= EPOLLIN | EPOLLPRI;
    }
    if (w) {
      pe.events |= EPOLLOUT;
    }
    if (r || w) {
      op = CONN_EV_NONE != oev ? EPOLL_CTL_MOD : EPOLL_CTL_ADD;
    } else {
      op = EPOLL_CTL_DEL;
    }

    if (-1 == epoll_ctl(s->ev_fd, op, fd, &pe)) {
      WARN("epoll_ctl() failed: %s", compat_strerror(errno));
    }
  }
#endif /* USE_EPOLL */

  c->set_eventmask(c, ev);
}

/**
 * This function is only used with kqueue() and because up to NetBSD 2.0H
 * closing a pipe doesn't cause a EVFILT_READ in a remote process reading
 * from the the pipe. Therefore, the other process would wait forever
 * instead of terminating automagically after the remote process.
 */
static int
server_watch_process(server_t *server, pid_t pid, server_process_cb_t cb)
#ifdef USE_KQUEUE
{
  server_private_t *s = &PRIV(server);
  
  static const struct timespec ts;
  struct kevent kev;

  EV_SET(&kev, pid, EVFILT_PROC, cb != NULL ? EV_ADD : EV_DELETE,
      0, NOTE_EXIT, PTR_TO_KEVENT_UDATA(cb));
  if (-1 == kevent(s->ev_fd, &kev, 1, NULL, 0, &ts)) {
    WARN("kevent() failed: %s", compat_strerror(errno));
    return -1;
  }
  return 0;
}
#else /* !USE_KQUEUE */
{
  (void) server;
  (void) pid;
  (void) cb;
  errno = ENOTSUP;
  return -1;
}
#endif /* USE_KQUEUE */

static inline int
server_periodic_handler(server_private_t *s, const struct timeval *tv)
{
  static struct timeval last_tv;
  
  if (!s->periodic_cb)
    return 0;

  if (!last_tv.tv_sec)
    last_tv = *tv;
    
  if (DIFFTIMEVAL(tv, &last_tv) / 1000 >= s->timeout) {
    s->periodic_cb((server_t *) s, tv);
    last_tv = *tv;
    return 1;
  }

  return 0;
}

typedef enum {
  POLL_READ = POLLIN | POLLRDNORM | POLLRDBAND | POLLPRI | POLLHUP,
  POLL_WRITE = POLLOUT | POLLWRNORM | POLLWRBAND,
  POLL_ANY = POLL_READ | POLL_WRITE
} poll_event_t;

static inline int
conn_ev_to_poll(connection_event_t ev)
{
  return  (ev & CONN_EV_READ ? POLL_READ : 0) |
          (ev & CONN_EV_WRITE ? POLL_WRITE : 0);
}

static void
server_create_pollset(server_private_t *s)
{
  size_t i, j;
 
  i = 0;
  for (j = 0; j < s->size_clients && i < s->num_clients; j++) {
    connection_t *c;
   
    c = s->clients[j];
    if (c) {
      struct pollfd *pfd, *base;

      assert(c != NULL);
      assert(i < s->num_clients);
      
      base = s->ev_arr;
      pfd = &base[i++];

      pfd->fd = c->get_fd(c);
      assert(pfd->fd >= 0);

      pfd->events = conn_ev_to_poll(c->get_eventmask(c));
    }
  }
}

static void
server_inform_clients_poll(server_private_t *s, const struct timeval *now)
{
  size_t i, j;

  i = 0;
  for (j = 0; j < s->size_clients && i < s->num_clients; j++) {
    connection_t *c;

    c = s->clients[j];
    if (c) {
      struct pollfd *pfd, *base;
      connection_event_t ev;

      assert(i < s->num_clients);
      base = s->ev_arr;
      pfd = &base[i++];

      if (c->get_fd(c) < 0) {
        /* This happens when an incoming connection was accept()ed and
         * the incoming cb handler was called which closed the
         * connection. */
        assert(-1 == c->get_fd(c));
        CRIT("Invalid file descriptor");
        s->num_closed++;
        continue;
      }

      if (c->get_fd(c) != pfd->fd) {
        CRIT("file descriptor mismatch; c->fd=%d; s->pfds[i].fd=%d",
            c->get_fd(c), pfd->fd);
        assert(c->get_fd(c) == pfd->fd);
      }

      if (0 == pfd->revents)
        continue;

      ev = CONN_EV_NONE;
      if (pfd->revents & POLL_READ)
        ev |= CONN_EV_READ;
      if (pfd->revents & POLL_WRITE)
        ev |= CONN_EV_WRITE;
      if (pfd->revents & POLLHUP)
        ev |= CONN_EV_HANGUP;
      if (pfd->revents & POLLERR)
        ev |= CONN_EV_ERROR;

      if (pfd->revents & POLLNVAL) {
        ev |= CONN_EV_ERROR;
        WARN("POLLNVAL (fd=%d)", pfd->fd);
      }

      if ((CONN_EV_ERROR & ev) || (ev & c->get_eventmask(c))) {
        c->event(c, ev, now);
        if (-1 == c->get_fd(c))
          s->num_closed++;
      }
    }
  }

  server_garbage_collect(s, now);
}

static int
server_add_client(server_t *server, connection_t *c)
{
  server_private_t *s = &PRIV(server);
  char *p;
  size_t i;
  
  assert(c != NULL);
  assert(c->get_fd(c) >= 0);

  p = s->ev_arr;
  if (s->num_clients == s->size_clients) {
    size_t n;

    n = s->size_clients;
    n = n < SERVER_CHUNKS ? SERVER_CHUNKS : n * 2;
    p = realloc(p, n * s->ev_size);
    if (!p)
      return -1;
  }
  assert(p != NULL);

  memset(&p[s->num_clients * s->ev_size], 0, s->ev_size);
  s->ev_arr = p;

  if (SERVER_TYPE_POLL == s->server_type) {
    struct pollfd *fds;

    fds = s->ev_arr;
    fds[s->num_clients].fd = c->get_fd(c);
    fds[s->num_clients].events = 0;
    fds[s->num_clients].revents = 0;
  }

  /*
   * This is OK when using poll() because the new client is appended
   * so it's not a problem when a event callback adds a client while
   * being in server_inform_clients().
   */
 
  if (s->num_clients == s->size_clients) {
    size_t n;
    
    n = s->size_clients;
    n = n < SERVER_CHUNKS ? SERVER_CHUNKS : n * 2;
    p = realloc(s->clients, n * sizeof s->clients[0]);
    if (!p)
      return -1;

    s->clients = cast_to_void_ptr(p);
    for (i = s->size_clients; i < n; i++) {
      s->clients[i] = NULL;
    }
    s->size_clients = n;
  }
  
  assert(s->num_clients < s->size_clients);
  for (i = 0; i < s->size_clients; i++) {
    if (!s->clients[i]) {
      s->clients[i] = c;
      break;
    }
  }
  assert(i < s->size_clients);
  s->num_clients++;

  return 0;
}

#ifdef USE_KQUEUE
static void
server_mainloop_kqueue(server_private_t *s)
{
  assert(s);
  assert(s->ev_fd != -1);

  for (;;) {
    int ret;

    ret = kevent(s->ev_fd, NULL, 0, s->ev_arr, s->num_clients, NULL);
    if (ret == -1) {
      if (EAGAIN != errno && EINTR != errno)
        WARN("kevent() failed: %s", compat_strerror(errno));
    } else if (ret > 0) {
      struct timeval now;
      bool periodic = false;
      size_t i;
      
      assert(ret >= 0 && (size_t) ret <= s->num_clients);
      compat_mono_time(&now);
      
      for (i = 0; i < (size_t) ret; i++) {
        const struct kevent *base = s->ev_arr, *kev = &base[i];
#if 0
         DBUG("kevent(): ident=%p filter=%lu flags=%lu fflags=%lu "
           "data=%" PRId64 " udata=%p %s",
           (void *) s->kevents[i].ident,
           (unsigned long) s->kevents[i].filter,
           (unsigned long) s->kevents[i].flags,
           (unsigned long) s->kevents[i].fflags,
           s->kevents[i].data,
           (void *) s->kevents[i].udata,
           s->kevents[i].flags & EV_EOF ? "EOF" : ""
         );
#endif

        if (EVFILT_TIMER == kev->filter) {
          assert(kev->ident == PTR2UINT(s));
          periodic = true;
        } else if (EVFILT_PROC == kev->filter) {
          server_process_cb_t cb;
          pid_t pid;

          pid = kev->ident;
          cb = (server_process_cb_t) KEVENT_UDATA_TO_PTR(kev->udata);
          assert(cb != NULL);
          cb((server_t *) s, pid);
        } else {
          int fd, n;
          connection_t *c;
          connection_event_t ev;

          fd = kev->ident;
          assert(fd >= 0);
          c = KEVENT_UDATA_TO_PTR(kev->udata);
          assert(c != NULL);
          assert(c->get_fd(c) == fd);
       
          ev = CONN_EV_NONE;
          n = 1;
          switch (kev->filter) {
          case EVFILT_READ:
            ev = CONN_EV_READ;
            if (c->get_listening(c)) {
              n = MIN(kev->data, 16);
            }
            if (0 == kev->data)
              ev |= CONN_EV_HANGUP;
            break;
          case EVFILT_WRITE:
            ev = CONN_EV_WRITE;
            if (0 == kev->data)
              ev |= CONN_EV_HANGUP;
            break;
          default:
            assert(0);
          }
        
          while (n-- > 0 && (ev & c->get_eventmask(c))) {
            c->event(c, ev, &now);
            fd = c->get_fd(c);
            if (-1 == fd) {
              s->num_closed++;
              break;
            }
          }

        }
 
      }

      server_garbage_collect(s, &now);
      
      if (periodic && s->periodic_cb) {
        s->periodic_cb((server_t *) s, &now);
      }
    
    } else if (0 == s->num_clients && 0 != s->timeout) {
      struct kevent kev;

      ret = kevent(s->ev_fd, NULL, 0, &kev, 1, NULL);
      switch (ret) {
      case 0: break;
      case -1:
        WARN("kevent() failed: %s", compat_strerror(errno));
        break;
      default:  
        if (EVFILT_TIMER == kev.filter) {
          assert(kev.ident == PTR2UINT(s));
          if (s->periodic_cb) {
            struct timeval tv;

            compat_mono_time(&tv);
            s->periodic_cb((server_t *) s, &tv);
          }
        }
      }
    }
  }
  
}

#endif /* USE_KQUEUE */

#ifdef USE_EPOLL
static bool
server_inform_client_epoll(connection_t *c, const struct epoll_event *e,
    const struct timeval *tv)
{
  connection_event_t ev;
  bool closed = false;

  assert(c);
  assert(e);
  assert(tv);

  ev = CONN_EV_NONE;
  if ((EPOLLIN | EPOLLPRI) & e->events)
    ev |= CONN_EV_READ;
  if (EPOLLOUT & e->events)
    ev |= CONN_EV_WRITE;
  if (EPOLLHUP & e->events)
    ev |= CONN_EV_HANGUP;
  if (EPOLLERR & e->events)
    ev |= CONN_EV_ERROR;

  if ((ev & (CONN_EV_ERROR | CONN_EV_HANGUP)) || (ev & c->get_eventmask(c))) {
    c->event(c, ev, tv);
    closed = -1 == c->get_fd(c);
  }

  return closed;
}

static void
server_inform_clients_epoll(server_t *server, int n, const struct timeval *tv)
{
  server_private_t *s;
  int i;
  
  assert(server);
  s = &PRIV(server);
  
  assert(n >= 0);
  assert((unsigned) n <= s->num_clients);
  assert(tv);
  
  for (i = 0; i < n; i++) {
    const struct epoll_event *pe, *base;
    connection_t *c;
    
    base = s->ev_arr;
    pe = &base[i];
    c = pe->data.ptr;
    assert(c);
    assert(-1 != c->get_fd(c));

    if (server_inform_client_epoll(c, pe, tv))
      s->num_closed++;
  }

  server_garbage_collect(s, tv);
}

static void
server_mainloop_epoll(server_private_t *s)
{
  assert(s);

  for (;;) {
    struct timeval tv;
    int ret;

    ret = epoll_wait(s->ev_fd, s->ev_arr, s->num_clients, s->timeout);
    if (-1 == ret && errno != EINTR && errno != EAGAIN) {
      WARN("epoll_wait() failed: %s", compat_strerror(errno));
      sleep(1); /* prevent exhausting CPU resources */
    }

    compat_mono_time(&tv);
    server_periodic_handler(s, &tv);
    if (ret > 0)
      server_inform_clients_epoll((server_t *) s, ret, &tv);
  }
}
#endif /* USE_EPOLL */

static void
server_mainloop_poll(server_private_t *s)
{
  assert(s);

  for (;;) {
    struct timeval tv;
    int ret;

    server_create_pollset(s);

    do {
      ret = poll(s->ev_arr, s->num_clients, s->timeout);
      if (-1 == ret && errno != EINTR && errno != EAGAIN) {
        WARN("poll failed: %s", compat_strerror(errno));
        sleep(1); /* prevent exhausting CPU resources */
      }

      compat_mono_time(&tv);
      if (server_periodic_handler(s, &tv))
        break;
    } while (ret == 0);
   
    if (ret > 0)
      server_inform_clients_poll(s, &tv);
  }
}

static void
server_mainloop(server_t *server)
{
  server_private_t *s;
  
  assert(server);
  s = &PRIV(server);

  set_signal(SIGPIPE, SIG_IGN);

  switch (s->server_type) {
  case SERVER_TYPE_KQUEUE:
#ifdef USE_KQUEUE
    server_mainloop_kqueue(s);
    return;
#else /* !USE_KQUEUE */
    break;
#endif /* USE_KQUEUE */

  case SERVER_TYPE_EPOLL:
#ifdef USE_EPOLL
    server_mainloop_epoll(s);
    return;
#else /* !USE_EPOLL */
    break;
#endif /* USE_EPOLL */

  case SERVER_TYPE_POLL:
    server_mainloop_poll(s);
    return;

  case SERVER_TYPE_NONE:
    break;
  }
  assert(0);
}

static int
server_error(server_t *s, const char *where, const char *what, int error)
{
  (void) s;
  fprintf(stderr, "%s: %s%s\n", where, what, compat_strerror(errno));
  return error;
}

static const struct server server_vtable = {
  &server_mainloop,
  &server_error,
  &server_add_client,
  &server_watch_client,
  &server_watch_process,
  &server_set_periodic_cb,
  &server_set_timeout,
  &server_destruct
};

static int
server_new_kqueue(server_private_t *s)
#ifdef USE_KQUEUE
{
  s->ev_fd = kqueue();
  if (-1 == s->ev_fd)
    return -1;

  s->ev_size = sizeof(struct kevent);
  s->server_type = SERVER_TYPE_KQUEUE;
  return 0;
}
#else /* !USE_KQUEUE */
{
  (void) s;
  errno = ENOTSUP;
  return -1;
}
#endif /* USE_KQUEUE */

static int
server_new_epoll(server_private_t *s)
#ifdef USE_EPOLL
{
  s->ev_fd = epoll_create(SERVER_CHUNKS);
  if (-1 == s->ev_fd)
    return -1;
  
  s->ev_size = sizeof(struct epoll_event);
  s->server_type = SERVER_TYPE_EPOLL;
  return 0;
}
#else /* !USE_EPOLL */
{
  (void) s;
  errno = ENOTSUP;
  return -1;
}
#endif /* USE_EPOLL */

static int
server_new_poll(server_private_t *s)
{
  s->ev_size = sizeof(struct pollfd);
  s->server_type = SERVER_TYPE_POLL;
  return 0;
}

server_t *
server_new(void)
{
  server_private_t *s = calloc(1, sizeof *s);

  if (s) {
    s->vtable = server_vtable;
    s->magic = server_magic;
    s->clients = NULL;
    s->size_clients = 0;
    s->num_clients = 0;
    s->periodic_cb = NULL;
    compat_mono_time(&s->last_gc);
    s->num_closed = 0;
   
    s->server_type = SERVER_TYPE_NONE;
    s->ev_arr = NULL;
    s->ev_size = 0;
    s->ev_fd = -1;

    if (
      0 != server_new_kqueue(s) &&
      0 != server_new_epoll(s) &&
      0 != server_new_poll(s)
    ) {
      free(s);
      return NULL;
    }
  }
  return (server_t *) s;
}


/* vi: set ai et sts=2 sw=2 cindent: */
